<?php

namespace InSegment\WorkersFramework\Queues;

use Shavshukov\EventAppBase\Model\ArrayModel;

class Worker
{
    /**
     * Name of the worker
     *
     * @var string
     */
    protected $workerName;
    
    /**
     * Name of the queue this worker consumes
     * 
     * @var string
     */
    protected $queueName;
    
    /**
     * Name of the queue for state update events for this worker
     * 
     * @var string
     */
    protected $serviceQueueName;
    
    /**
     * Name of the queue for not acquired messages
     *
     * @var string
     */
    protected $dlxQueueName;
    
    /**
     * Whether to return messages from dlx queue to worker queue
     *
     * @var bool
     */
    protected $dlxRequeue;
    
    /**
     * TTL of messages in dlx queue
     *
     * @var int
     */
    protected $dlxTtl;
    
    /**
     * Worker will prefer jobs that he attracts
     *
     * @var array
     */
    protected $attracts;
    
    /**
     * Worker will not prefer jobs that ha avoids
     *
     * @var array
     */
    protected $avoids;
    
    /**
     * Status of the worker
     *
     * @var \Shavshukov\EventAppBase\Model\ArrayModel 
     */
    protected $status;
    
    /**
     * Nodes this worker works on
     *
     * @var array|null
     */
    protected $nodes;
    
    /**
     * Whether the worker works on the current node
     *
     * @var bool
     */
    protected $presentOnNode;
    
    /**
     * Constructor
     * 
     * @param string $workerName
     * @param array $config
     * @param string $nodeName
     */
    public function __construct(string $workerName, array $config, string $nodeName)
    {
        $this->workerName = $workerName;
        $this->queueName = $config['work_queue'];
        $this->serviceQueueName = $config['service_queue'];
        
        if (!$this->queueName || !$this->serviceQueueName) {
            throw new \Exception("A worker configuration must have 'work_queue' and 'service_queue'!");
        }
        
        $this->dlxQueueName = $config['dlx_queue'] ?? null;
        $this->dlxRequeue = $config['dlx_requeue'] ?? true;
        $this->dlxTtl = $config['dlx_ttl'] ?? null;
        
        $this->attracts = $config['attracts'] ?? [];
        $this->avoids = $config['avoids'] ?? [];
        $this->jobWeight = $config['job_weight'] ?? 1;
        $this->status = ArrayModel::from([
            'active' => false,
            'workerState' => 'offline',
            'startTime' => null
        ]);
        
        if (isset($config['nodes'])) {
            if (!is_array($config['nodes'])) {
                throw new \Exception("Worker 'nodes' must be an array, it is not in worker '{$workerName}'");
            }
            
            $this->nodes = $config['nodes'];
            $this->presentOnNode = in_array($nodeName, $config['nodes']);
        } else {
            $this->presentOnNode = true;
        }
    }
    
    /**
     * Get structure to declare worker queues
     * 
     * @param \InSegment\WorkersFramework\Queues\ConnectionsList $connections
     * @return \InSegment\WorkersFramework\Queues\WorkerDeclareStructure
     * @throws \Exception
     */
    public function getDeclareStructure(ConnectionsList $connections): WorkerDeclareStructure
    {
        $declareWork = $connections->declareQueue($this->queueName);
        $declareService = $connections->declareQueue($this->serviceQueueName);
        
        if ($this->dlxQueueName) {
            if ($connections->getQueue($this->queueName)->getConnection() !== $connections->getQueue($this->dlxQueueName)->getConnection()) {
                throw new \Exception("Work queue and DLX queue must be on the same connection!");
            }

            $declareDlx = $connections->declareQueue($this->dlxQueueName);
            $declareWork->setDlxQueueName($this->dlxQueueName);
            
            if ($this->dlxRequeue) {
                $declareDlx->setDlxQueueName($this->queueName);
                
                if ($this->dlxTtl) {
                    $declareDlx->setMessageTtl($this->dlxTtl);
                }
            }
        } else {
            $declareDlx = null;
        }
        
        return new WorkerDeclareStructure($declareWork, $declareService, $declareDlx);
    }
    
    /**
     * Get Consumer status as array
     * 
     * @return array
     */
    public function getStatus(): array
    {
        return [
            'worker_state' => $this->status->get('workerState'),
            'active' => $this->status->get('active'),
        ];
    }
    
    /**
     * Whether there is currently a process under that worker
     * 
     * @return bool
     */
    public function isActive(): bool
    {
        return $this->status->get('active');
    }
    
    /**
     * Change values in status of the worker according to updates from the process
     * 
     * @param array $modification
     * @return null
     */
    public function modifyStatus(array $modification)
    {
        $this->status->modify($modification);
    }
    
    /**
     * Get name of the worker
     * 
     * @return string
     */
    public function getName(): string
    {
        return $this->workerName;
    }
    
    /**
     * Get name of the queue this worker consumes
     * 
     * @return string
     */
    public function getQueueName(): string
    {
        return $this->queueName;
    }
    
    /**
     * Get name of the queue for state update events for this worker
     * 
     * @return string
     */
    public function getServiceQueueName(): string
    {
        return $this->serviceQueueName;
    }
    
    /**
     * Get jobs that this worker attracts
     * 
     * @return array
     */
    public function getAttracts(): array
    {
        return $this->attracts;
    }
    
    /**
     * Get jobs that this worker avoids
     * 
     * @return array
     */
    public function getAvoids(): array
    {
        return $this->avoids;
    }
    
    /**
     * Whether the worker present on the current node
     * 
     * @return bool
     */
    public function isPresentOnNode(): bool
    {
        return $this->presentOnNode;
    }
    
    /**
     * Get whether the worker has common nodes with specified
     * 
     * @param array $nodes
     * @return bool
     */
    public function hasNodesIntersect(array $nodes): bool
    {
        return $this->nodes === null || count(array_intersect($this->nodes, $nodes)) > 0;
    }
    
}

