<?php

namespace InSegment\WorkersFramework\Services;

use Shavshukov\EventAppBase\ApplicationInterface;
use React\ChildProcess\Process;
use React\Filesystem\FilesystemInterface;
use React\Filesystem\Filesystem;
use Carbon\Carbon;

class WorkerGc
{
    /**
     * Do GC of the lock file every N stopped processes
     * 
     * @var int
     */
    const GC_INTERVAL = 100;

    /**
     * Application instance
     * 
     * @var \Shavshukov\EventAppBase\ApplicationInterface $app 
     */
    protected $app;

    /**
     * Full path to the workers lock file
     * 
     * @var string
     */
    protected string $lockfile;

    /**
     * The non-blocking interface to work with files
     * 
     * @var FilesystemInterface
     */
    protected FilesystemInterface $filesystem;

    /**
     * Started processes
     * 
     * @var array [
     *      int $PID => \React\ChildProcess\Process $process,
     * ]
     */
    protected static array $startedProcesses;

    /**
     * Number of the child processes closed
     * 
     * @var int
     */
    protected static int $closedProcessesCounter = 0;

    /**
     * GC loop timer should be sarted only once
     * 
     * @var bool
     */
    protected static bool $gcStarted = false;

    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\ApplicationInterface $app
     */
    public function __construct(ApplicationInterface $app)
    {
        $this->app = $app;
        $this->filesystem = Filesystem::create($this->app->getLoop());
        $this->lockfile = $this->app->config('worker_pid_lockfile');
        if ($this->lockfile && !self::$gcStarted) {
            $this->app->getLoop()->addPeriodicTimer(1, function () {
                $this->workerLockfileGc();
            });
            echo "Workers PID lockfile GC started\n";
            self::$gcStarted = true;
        }
    }

    public function register(\React\ChildProcess\Process $process)
    {
        if (!$this->lockfile) {
            return;
        }

        $childPid = $process->getPid();
        echo "Spawning child process PID {$childPid}\n";
        $this->filesystem->file($this->lockfile)->open('w')->then(function ($stream) use ($childPid) {
            $stream->end("{$childPid}\n");
        });

        self::$startedProcesses[$childPid] = $process;
    }

    protected function workerLockfileGc()
    {
        foreach (self::$startedProcesses as $pid => $process) {
            if ($process->isStopped() || $process->isTerminated() || !$process->isRunning()) {
                unset(self::$startedProcesses[$pid]);
                ++self::$closedProcessesCounter;
            }
        }

        if (self::$closedProcessesCounter !== self::GC_INTERVAL) {
            return;
        }
        self::$closedProcessesCounter = 0;

        echo "GC workers PID lockfile...\n";
        $handler = fopen($this->lockfile, 'r+');
        if (flock($handler, LOCK_EX)) {
            ftruncate($handler, 0);
            fwrite($handler, implode("\n", array_keys(self::$startedProcesses)));
            fflush($handler);
            flock($handler, LOCK_UN);
        } else {
            throw new \Exception(__METHOD__ . ": can not lock file {$this->lockfile} for garbage collecting");
        }
        fclose($handler);

        echo "GC success\n";
    }

    public function killPreviousPids()
    {
        $lockfile = $this->lockfile;
        if (!$lockfile || !file_exists($lockfile) || !is_readable($lockfile)) {
            echo "Warning: workers PID lock file does not exist!\n";
            return;
        }
        if (!preg_match('wtmp begins (.+)', shell_exec('last reboot -1'), $matches)) {
            echo "Warning: can not get the last reboot time!\n";
            return;
        }

        try {
            $rebootDateTime = Carbon::parse($matches[1]);
            $fileDateTime = Carbon::createFromTimestamp(filectime($lockfile));
        } catch (\Exception $e) {
            echo "Warning: can not parse the dates: {$e->getTraceAsString()}\n";
            return;
        }

        if ($rebootDateTime->greaterThan($fileDateTime)) {
            echo "Deleting the lock file after system reboot\n";
            unlink($lockfile);
            return;
        }

        if (($handle = fopen($lockfile, 'r'))) {
            while (($pid = fgets($handle)) !== false) {
                echo "Killing process PID {$pid}";
                (new Process("kill {$pid}"))->start($this->app->getLoop());
            }
            fclose($handle);
            unlink($lockfile);
        }
    }
}
